# Attempt to use the built-in PostgreSQL json field, but if that is not
# available try the 3rd party jsonfield one.
try:
    from django.contrib.postgres.fields import JSONField as BaseJSONField
except:
    from jsonfield.fields import JSONField as BaseJSONField


class JSONField(BaseJSONField):
    pass



from django.apps import AppConfig


class PapertrailConfig(AppConfig):
    name = 'papertrail'

    def ready(self):
        import papertrail
        from papertrail import models

        for func in (
            'log',
            'related_to',
            'objects_not_represented',
            'objects_represented',
            'filter',
            'exclude',
            'all_types',
        ):
            setattr(papertrail, func, getattr(models, func))



from django import template
from django.core.urlresolvers import NoReverseMatch, reverse
from django.db import models

register = template.Library()


@register.filter
def adminview(value, view='change'):
    '''
    Given an instance (`value`), generate a link to the specified admin `view`.

    Example:
        {{ user|adminview:'history' }} -> /admin/user/{user_id}/history/
    '''
    if isinstance(value, models.Model):
        try:
            change_form_url = reverse(
                'admin:{app}_{module}_{view}'.format(
                    app=value._meta.app_label,
                    module=value._meta.model_name,
                    view=view
                ),
                args=[value.pk]
            )
            return change_form_url
        except NoReverseMatch:
            pass
    return None


@register.filter
def has_papertrail(model_instance):
    url = adminview(model_instance, view='papertrail')
    return url is not None



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('papertrail', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='entry',
            options={'ordering': ['-timestamp'], 'get_latest_by': 'timestamp', 'verbose_name_plural': 'entries'},
        ),
    ]



import collections
import itertools
import json

from django.conf.urls import url
from django.contrib import admin
from django.core import serializers
from django.core.urlresolvers import reverse
from django.db.models import Q
from django.shortcuts import get_object_or_404, render
from django.utils.encoding import force_text
from django.utils.html import format_html
from django.utils.text import capfirst
from django.utils.translation import ugettext as _

import papertrail
from papertrail.models import Entry, EntryRelatedObject


def admin_reverse_for_model(model):
    return reverse('admin:{}_{}_change'.format(model._meta.app_label, model._meta.model_name), args=[model.id])


class AdminEventLoggerMixin(object):
    '''
    Mixin for ModelAdmin classes to log admin actions to the papertrail
    application as well as to Django's built-in admin logging.
    '''
    papertrail_object_template = None

    # A map of category -> events used by the papertrail viewer to render
    # display filters
    papertrail_type_filters = {}

    # Override this field to change the 'target' object used for retrieving and
    # displaying history.  It should be a double__underscore__delimited
    # property related to the primary model admin object.
    papertrail_field = None

    def _record_changes(self, obj, fields=None):
        '''
        Records the state of `obj` to a JSON-serializable object, optionally
        recording only values in a list of `fields`.  If `fields` is not
        specified, all fields will be recorded.
        '''
        rec = json.loads(serializers.serialize('json', [obj]))[0]
        if fields:
            rec['fields'] = {k: v for k, v in rec['fields'].items()
                             if k in fields}
        return rec

    def _resolve_pt_object(self, object):
        instance = object

        if self.papertrail_field:
            for field_name in self.papertrail_field.split('__'):
                instance = getattr(instance, field_name, None)
                if not instance:
                    return object

        return instance

    def log_addition(self, request, object, message=None):
        if message is not None:
            super(AdminEventLoggerMixin, self).log_addition(request, object, message)
            try:
                message = json.loads(message)
            except ValueError:
                pass
        else:
            super(AdminEventLoggerMixin, self).log_addition(request, object)

        fields = self._record_changes(object)['fields']
        return papertrail.log(
            'admin-edit',
            'Created object',
            data={
                'action': 'add',
                'fields': fields,
                'message': message,
            },
            targets={
               'acting_user': request.user,
               'instance': self._resolve_pt_object(object),
            },
        )

    def log_change(self, request, object, message):
        super(AdminEventLoggerMixin, self).log_change(request, object, message)

        # construct_change_message() creates a JSON message that we load and
        # store here. (In case we don't get JSON back for some reason, still
        # store the message)
        try:
            data = {'changes': json.loads(message), 'action': 'change'}
        except ValueError:
            data = {'message': message, 'action': 'change'}

        return papertrail.log(
            'admin-edit',
            'Updated object',
            data=data,
            targets={
                'acting_user': request.user,
                'instance': self._resolve_pt_object(object),
            },
        )

    def log_deletion(self, request, object, object_repr):
        super(AdminEventLoggerMixin, self).log_deletion(request, object, object_repr)

        fields = self._record_changes(object)['fields']
        return papertrail.log(
            'admin-edit',
            'Deleted object',
            data={
                'action': 'delete',
                'fields': fields,
            },
            targets={
                'acting_user': request.user,
                'instance': self._resolve_pt_object(object),
            },
        )

    def get_urls(self):
        info = self.model._meta.app_label, self.model._meta.model_name

        urlpatterns = [
            url(r'^(.+)/papertrail/', self.admin_site.admin_view(self.view_papertrail_item), name=u'{0}_{1}_papertrail'.format(*info)),
        ] + super(AdminEventLoggerMixin, self).get_urls()

        return urlpatterns

    def get_actions(self, request):
        actions = super(AdminEventLoggerMixin, self).get_actions(request)
        func, name, desc = self.get_action('view_papertrail')
        actions[name] = (func, name, desc)
        return actions

    def view_papertrail_item(self, request, object_id, extra_context=None):
        get_object_or_404(self.model, id=object_id)
        return self.view_papertrail(request, self.model.objects.filter(id=object_id))

    def view_papertrail(self, request, queryset, extra_context=None):
        '''
        Action for displaying the papertrail for one or more selected admin
        items.
        '''

        # Map to alternate queryset if specified
        if self.papertrail_field:
            queryset = self._map_to_related_queryset(queryset, self.papertrail_field)

        action_list = Entry.objects.related_to(queryset).select_related()
        opts = queryset.model._meta
        app_label = opts.app_label

        available_type_filters = collections.OrderedDict(self.papertrail_type_filters)
        if available_type_filters:
            available_type_filters['Other'] = ('__other', )
        selected_type_filters = request.POST.getlist('papertrail-selected-type-filters')
        selected_types = list(itertools.chain.from_iterable(available_type_filters.get(k, []) for k in selected_type_filters))
        if '__other' in selected_types:
            all_types = list(itertools.chain.from_iterable(available_type_filters.values()))
            action_list = action_list.filter(Q(type__in=selected_types) | ~Q(type__in=all_types))
        elif selected_types:
            action_list = action_list.filter(type__in=selected_types)

        if queryset.count() == 1:
            obj = queryset[0]
            title = _('Paper Trail: %s') % force_text(obj)
        else:
            obj = None
            title = _('Paper Trail: %s %s') % (queryset.count(), opts.verbose_name_plural)

        context = {
            'title': title,
            'action_list': action_list,
            'module_name': capfirst(force_text(opts.verbose_name_plural)),
            'app_label': app_label,
            'opts': opts,
            'object': obj,
            'available_type_filters': available_type_filters,
            'selected_type_filters': selected_type_filters,
            'selected_actions': request.POST.getlist('_selected_action'),
        }
        context.update(extra_context or {})

        return render(request, self.papertrail_object_template or [
            "admin/%s/%s/object_papertrail.html" % (app_label, opts.object_name.lower()),
            "admin/%s/object_papertrail.html" % app_label,
            "admin/object_papertrail.html"
        ], context)
    view_papertrail.short_description = 'View Paper Trail'

    def _map_to_related_queryset(self, queryset, field):
        '''
        Given a queryset and an double__underscore__delimited field relation,
        return a list of the objects that are the target of that relation for
        all objects in the queryset.
        '''
        model = queryset.model

        # The easy part is finding the pks of the related objects
        pks = queryset.values_list('{}__pk'.format(field), flat=True)

        # The hard part is traversing the relation string and finding out what
        # model we're actually looking for.
        segments = field.split('__')
        for segment in segments:
            field = getattr(model, segment).field
            try:
                model = field.related.parent_model
            except AttributeError:
                model = field.rel.model

        # Once we have both pieces, we can just query the model for the ids
        return model.objects.filter(pk__in=pks)

    def construct_change_message(self, request, form, formsets, add=False):
        '''
        Construct a detailed change message from a changed object, including
        related objects updated via subforms.  Returns a JSON string containing
        a structure detailing the fields changed and their updated values.
        '''
        def add_related_change(changes, obj, action='change', fields=None):
            rec = self._record_changes(obj, fields=fields)
            rec['action'] = action
            changes['related_changes'].append(rec)
            return rec

        changes = {
            'action': 'change',
            'fields': self._record_changes(form.instance, form.changed_data)['fields'],
            'related_changes': [],
        }

        for formset in (formsets or []):
            for obj in formset.new_objects:
                add_related_change(changes, obj, action='add')
            for obj, changed_fields in formset.changed_objects:
                add_related_change(changes, obj, action='change', fields=changed_fields)
            for obj in formset.deleted_objects:
                add_related_change(changes, obj, action='add')

        return json.dumps(changes)


class EntryRelatedObjectInline(admin.StackedInline):
    model = EntryRelatedObject
    extra = 0
    fields = ('relation_name', 'related_content_type', 'related_model', )
    readonly_fields = ('related_model', )

    def related_model(self, obj):
        related_obj = obj.related_object
        return format_html(
            u'<a href="{}">{}</a>',
            admin_reverse_for_model(related_obj),
            related_obj,
        )


@admin.register(Entry)
class EntryAdmin(admin.ModelAdmin):
    list_display = ('type', 'message', 'timestamp', )
    search_fields = ('type', )
    inlines = (EntryRelatedObjectInline, )



import unittest
from django import VERSION as django_version
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.dispatch import receiver
from django.test import TestCase
from django.utils import timezone
from papertrail import log, signals
from papertrail.models import Entry, related_to_q, replace_object_in_papertrail

User = get_user_model()


class TestBasic(TestCase):

    def test_entry_logging(self):

        log('test', 'Testing entry')

        user = User.objects.create_user('testuser', 'test@example.com')
        log('test-user-created', 'User created', targets={'user': user})

        group = Group.objects.create(name='Test Group')
        log('test-group-created', 'Group created', targets={'group': group})

        group.user_set.add(user)
        log('test-group-added-user', 'User added to group', targets={'user': user, 'group': group})

        log('test-extra-data', 'Testing extra data', data={'key': 'value'})

        tznow = timezone.now()
        overridden_timestamp_entry = log(
            'test-overridden-timestamp',
            'Testing overriding a timestamp for operations like importing',
            timestamp=tznow,
        )
        self.assertEqual(tznow, overridden_timestamp_entry.timestamp)

        qs = Entry.objects.filter(type__startswith='test')
        self.assertEqual(qs.count(), 6)

        self.assertEqual(qs.related_to(user).count(), 2)
        self.assertEqual(qs.related_to(group).count(), 2)
        self.assertEqual(qs.related_to(user=user).count(), 2)
        self.assertEqual(qs.related_to(group=group).count(), 2)
        self.assertEqual(qs.related_to(user=group).count(), 0)
        self.assertEqual(qs.related_to(group=user).count(), 0)
        self.assertEqual(qs.related_to(user, group).count(), 1)
        self.assertEqual(qs.related_to(user=user, group=group).count(), 1)

        # test chaining and equivalence
        self.assertEqual(set(qs.related_to(user=user)
                               .related_to(group=group)
                               .values_list('id', flat=True)),
                         set(qs.related_to(user=user, group=group)
                               .values_list('id', flat=True)))

        # test related_to Q object
        self.assertEqual(set(qs.related_to(user, group)
                               .values_list('id', flat=True)),
                         set(qs.filter(related_to_q(user))
                               .filter(related_to_q(group))
                               .values_list('id', flat=True)))

        self.assertEqual(qs.filter(related_to_q(user) | related_to_q(group))
                           .distinct().count(),
                         3)

    def test_signals(self):
        event_logged_counter = [0]

        @receiver(signals.event_logged)
        def on_event_logged(sender, **kwargs):
            event_logged_counter[0] += 1

        log('test', 'Testing signal')
        self.assertEqual(event_logged_counter[0], 1)

    def test_setters_and_getters(self):
        e = log('test-entry', 'Test Entry')
        self.assertEqual(e.targets_map, {})

        user = User.objects.create_user('testuser', 'test@example.com')
        e.set('target1', user)

        # Basic lookup
        self.assertEqual(e.get('target1'), user)
        self.assertEqual(e['target1'], user)

        # Invalid key lookup and getting a default for non-existent targets
        with self.assertRaises(KeyError):
            e['target2']
        self.assertEqual(e.get('target2'), None)
        self.assertEqual(e.get('target2', 'a-default-value'), 'a-default-value')

        # Contains ('in') implementation
        self.assertTrue('target1' in e)
        self.assertFalse('target2' in e)

        # Setting and retrieving a 'virtual' target
        user_type = ContentType.objects.get_for_model(User)
        e.set('virtual', (user_type, 10000))
        self.assertEqual(e['virtual'], None)
        self.assertEqual(e.get('virtual', 'a-default-value'), None)

    def test_change_related_object(self):
        user1 = User.objects.create_user('testuser1', 'test1@example.com')
        user2 = User.objects.create_user('testuser2', 'test2@example.com')

        Entry.objects.all().delete()

        log('test-entry', 'This is a test entry', targets={'user': user1})
        log('test-entry', 'This is a test entry', targets={'user': user2})

        qs = Entry.objects.all()
        self.assertEqual(qs.related_to(user1).count(), 1)
        self.assertEqual(qs.related_to(user2).count(), 1)

        replace_object_in_papertrail(user1, user2)

        qs = Entry.objects.all()
        self.assertEqual(qs.related_to(user1).count(), 0)
        self.assertEqual(qs.related_to(user2).count(), 2)

    def test_objects_represented(self):
        user1 = User.objects.create_user('testuser1', 'test1@example.com')
        user2 = User.objects.create_user('testuser2', 'test2@example.com')
        user3 = User.objects.create_user('testuser3', 'test3@example.com')
        some_group = Group.objects.create(name='some group')
        some_group2 = Group.objects.create(name='some group 2')
        some_group3 = Group.objects.create(name='some group 3')

        log('test-entry', 'Test 1', targets={'user': user1, 'group': some_group})
        log('test-entry', 'Test 2', targets={'user': user2, 'group': some_group})
        log('test-entry', 'Test 3', targets={'user': user2, 'group': some_group2})
        log('test-entry', 'Test 4', targets={'blah': user1, 'group': some_group3})

        all_users = User.objects.all()
        self.assertEqual(
            set(Entry.objects.filter(type='test-entry').objects_represented(all_users, 'user')),
            set([user1, user2])
        )
        self.assertEqual(
            set(Entry.objects.filter(type='test-entry').related_to(group=some_group).objects_represented(all_users, 'user')),
            set([user1, user2]),
        )
        self.assertEqual(
            Entry.objects.filter(type='test-entry').related_to(group=some_group2).objects_represented(all_users, 'user').get(),
            user2,
        )
        self.assertTrue(
            Entry.objects.filter(type='test-entry').related_to(group=some_group3).exists(),
        )
        self.assertFalse(
            Entry.objects.filter(type='test-entry').related_to(group=some_group3).objects_represented(all_users, 'user').exists(),
        )
        self.assertEqual(
            Entry.objects.filter(type='test-entry').objects_not_represented(all_users, 'user').get(),
            user3
        )

    @unittest.skipIf(django_version < (1, 9), 'no native JSONField')
    def test_json_field_filtering(self):
        log('test', 'test', data={
            'field': 'value',
        })

        self.assertTrue(
            Entry.objects.filter(data__field='value').exists()
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import papertrail.fields


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Entry',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('timestamp', models.DateTimeField(db_index=True)),
                ('type', models.CharField(max_length=50, db_index=True)),
                ('message', models.CharField(max_length=512)),
                ('data', papertrail.fields.JSONField(null=True)),
                ('external_key', models.CharField(max_length=255, null=True, db_index=True)),
            ],
            options={
                'ordering': ['-timestamp'],
                'get_latest_by': 'timestamp',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='EntryRelatedObject',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('relation_name', models.CharField(max_length=100, db_index=True)),
                ('related_id', models.PositiveIntegerField(db_index=True)),
                ('entry', models.ForeignKey(related_name='targets', to='papertrail.Entry')),
                ('related_content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



from __future__ import print_function

import datetime
import json
import logging
from decimal import Decimal

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.utils import timezone

from papertrail import signals
from papertrail.fields import JSONField

LOG = logging.getLogger(__name__)


def coerce_to_queryset(instance_or_queryset):
    if isinstance(instance_or_queryset, models.Model):
        instance = instance_or_queryset
        return instance.__class__.objects.filter(pk=instance.pk)
    else:
        return instance_or_queryset


def related_to_q(obj, relation_name=None):
    '''
    Create a Q object expressing an event relation with an optional name.
    This is useful as a building block for Entry.objects.related_to(), and
    can be used to provide better query control for more complex queries
    without the boilerplate of directly querying an Entry's related objects.

    Example 1: OR query

        Entry.objects.filter(related_to(user1) | related_to(user2)).distinct()

    Example 2: building block to Entry.objects.related_to()

        The following are equivalent:

        Entry.objects.related_to(user=user1, group=group1)
        Entry.objects.filter(related_to_q(user1, 'user'))
                     .filter(related_to_q(group1, 'group'))

    '''
    related_qs = coerce_to_queryset(obj)
    content_type = ContentType.objects.get_for_model(related_qs.model)
    filters = {
        'targets__related_content_type': content_type,
        'targets__related_id__in': related_qs,
    }
    if relation_name:
        filters.update({'targets__relation_name': relation_name})
    return models.Q(**filters)


class RelatedObjectsQuerySetMixin(object):
    def _get_object_ids_in_papertrail(self, qs, relation_name):
        '''
        A helper method that receives an arbitrary queryset `qs` and a relation name
        `relation_name` and returns a list of object ids that are part of that queryset
        and ARE referenced by the `self` entries queryset.
        relation_name needs to point to the same type of objects the queryset this method
        receives is pointing to.

        For example, if qs has User objects whose ids are 1, 2, 3 and `self` has only
        one papertrail entry and that entry has targets={'user': user2}, calling
        _get_object_ids_in_papertrail(qs, 'user') would return only the id of user 2
        (because user1 and user3 do not have papertrail entries pointing at them)
        '''
        # Filter ourselves to only entries relating to the queryset the user
        # is passing us
        entries_related_to_qs = self.related_to(**{relation_name: qs})

        # Get the RelatedObject model (this would typically be EntryRelatedObject)
        targets_model = self.model().targets_model

        # Query related objects to get IDs of related objects that satisfy the following conditions:
        related_entries = (
            targets_model.objects
            # 1) They point to entries that are related to the queryset we received
            .filter(**{targets_model.owner_field_name + '__in': entries_related_to_qs})
            # 2) Their relation_name matches the one the user is querying for
            .filter(relation_name=relation_name)
        )

        # Return
        return related_entries.values_list('related_id', flat=True)

    def objects_not_represented(self, qs, relation_name):
        return qs.exclude(id__in=self._get_object_ids_in_papertrail(qs, relation_name))

    def objects_represented(self, qs, relation_name):
        return qs.filter(id__in=self._get_object_ids_in_papertrail(qs, relation_name))

    def related_to(self, *relations, **named_relations):
        '''
        Filter entries based on objects they pertain to, either generically or
        by a specific relation type.  If multiple relations are specified, the
        filter combines them with AND semantics.

        Examples:

            Tracking a simple 'follow' event where one user follows another,
            which is logged as:

            > user1 = User.objects.get(...)
            > user2 = User.objects.get(...)
            > log('user-followed', 'User followed another user', targets={'follower': user1, 'following': user2})

            First, a simple query for all events for user1, regardless of
            the type of relationship:

            > Entry.objects.related_to(user1)

            Next, to query for events involving both user1 and user2.

            > Entry.objects.related_to(user1, user2)

            Finally, to query for specific relationships, such as user1
            following user2:

            > Entry.objects.related_to(follower=user1, following=user2)
        '''
        entry_qs = self
        all_relations = [(None, r) for r in relations] + list(named_relations.items())

        for name, relation in all_relations:
            entry_qs = entry_qs.filter(related_to_q(relation, name))

        return entry_qs.distinct()


class EntryQuerySet(models.query.QuerySet, RelatedObjectsQuerySetMixin):
    def all_types(self):
        return self.order_by('type').values_list('type', flat=True).distinct()


class ModelWithRelatedObjectsMixin(object):
    def get(self, target, default=None):
        try:
            return self[target]
        except KeyError:
            return default

    def set(self, target_name, val, replace=True):
        '''
        Sets the updated target value, optionally replacing the existing target
        by that name.  If `replace` is False, raises an error if the target
        already exists.  `val` is generally a Django model instance, but can
        also be a tuple of (content_type, id) to reference an object as the
        contenttypes app does (this also allows references to deleted objects).
        '''
        target = self.get(target_name)
        if target and not replace:
            raise ValueError('Target {} already exists for this event'.format(target_name))

        attributes = {
            self.targets_model.owner_field_name: self,
            'relation_name': target_name,
        }
        target = target or self.targets_model(**attributes)
        if type(val) == tuple:
            content_type, object_id = val
            target.related_content_type = content_type
            target.related_id = object_id
            target.save()
        elif val:
            target.related_object = val
            target.save()
        return target

    @property
    def targets_map(self):
        return dict((t.relation_name, t.related_object) for t in self.targets.all())

    def update(self, targets_map):
        for target, val in (targets_map or {}).items():
            self[target] = val

    def __getitem__(self, target_name):
        # Optimization, in case we pre-fetched targets
        if hasattr(self, '_prefetched_objects_cache') and 'targets' in self._prefetched_objects_cache:
            for target in self._prefetched_objects_cache['targets']:
                if target.relation_name == target_name:
                    return target.related_object
            raise KeyError

        try:
            target = self.targets.get(relation_name=target_name)
            return target.related_object
        except self.targets_model.DoesNotExist:
            raise KeyError

    def __setitem__(self, target, val):
        return self.set(target, val)

    def __contains__(self, target_name):
        return self.targets.filter(relation_name=target_name).exists()


class Entry(models.Model, ModelWithRelatedObjectsMixin):
    timestamp = models.DateTimeField(db_index=True)
    type = models.CharField(max_length=50, db_index=True)
    message = models.CharField(max_length=512)
    data = JSONField(null=True)

    # Field for storing a custom 'key' for looking up specific events from
    # external sources.  This can be used to quickly and precisely look up
    # events that you can derive natural keys for, but aren't easily queryable
    # using Entry's other fields.
    external_key = models.CharField(max_length=255, null=True, db_index=True)

    objects = EntryQuerySet.as_manager()

    def __unicode__(self):
        return u'{} - {}'.format(self.type, self.message)

    class Meta:
        ordering = ['-timestamp']
        get_latest_by = 'timestamp'
        verbose_name_plural = 'entries'

    @property
    def targets_model(self):
        return EntryRelatedObject


# If you want to subclass this, make sure that the foreign key
# related name is called targets

class RelatedObject(models.Model):
    relation_name = models.CharField(max_length=100, db_index=True)
    related_content_type = models.ForeignKey(ContentType)
    related_id = models.PositiveIntegerField(db_index=True)
    related_object = GenericForeignKey('related_content_type', 'related_id')

    class Meta:
        abstract = True

    @property
    def safe_related_object(self):
        try:
            return self.related_object
        # This will happen if the model class of this object is not available, for example if the app providing it has been removed from INSTALLED_APPS
        except AttributeError:
            return None


class EntryRelatedObject(RelatedObject):
    entry = models.ForeignKey('Entry', related_name='targets')
    owner_field_name = 'entry'


def replace_object_in_papertrail(old_obj, new_obj, entry_qs=None):
    entry_qs = entry_qs or Entry.objects.all()
    old_obj_type = ContentType.objects.get_for_model(old_obj.__class__)
    new_obj_type = ContentType.objects.get_for_model(new_obj.__class__)
    related_qs = (EntryRelatedObject.objects.filter(
        entry__in=entry_qs,
        related_content_type=old_obj_type,
        related_id=old_obj.pk
        ))
    related_qs.update(related_content_type=new_obj_type,
                      related_id=new_obj.pk)


def json_default(o):
    """
    A `default` method for allowing objects with dates/decimals to be encoded into JSON.
    Usage: json.dumps(obj, default=_json_default)
    """
    if hasattr(o, 'to_json'):
        return o.to_json()
    if isinstance(o, Decimal):
        return str(o)
    if isinstance(o, datetime.datetime):
        if o.tzinfo:
            return o.strftime('%Y-%m-%dT%H:%M:%S%z')
        return o.strftime('%Y-%m-%dT%H:%M:%S')
    if isinstance(o, datetime.date):
        return o.strftime('%Y-%m-%d')
    if isinstance(o, datetime.time):
        if o.tzinfo:
            return o.strftime('%H:%M:%S%z')
        return o.strftime('%H:%M:%S')


def json_serializeable(obj):
    """
    Attempts to return a copy of `obj` that is JSON serializeable
    """
    return json.loads(json.dumps(obj, default=json_default))


def log(event_type, message, data=None, timestamp=None, targets=None, external_key=None, data_adapter=json_serializeable):
    data_adapter = data_adapter or (lambda obj: obj)
    try:
        timestamp = timestamp or timezone.now()
        with transaction.atomic():

            # Enforce uniqueness on event_type/external_id if an external id is
            # provided
            if external_key:
                entry, created = Entry.objects.get_or_create(
                    type=event_type,
                    external_key=external_key,
                    defaults={
                        'message': message,
                        'data': data,
                        'timestamp': timestamp,
                    }
                )
                if not created:
                    return
            else:
                entry = Entry.objects.create(
                        type=event_type,
                        message=message,
                        data=data_adapter(data),
                        timestamp=timestamp,
                    )

            entry.update(targets)
            if getattr(settings, 'PAPERTRAIL_SHOW', False):
                WARNING = u'\033[95m'
                ENDC = u'\033[0m'
                print(WARNING + u'papertrail ' + ENDC + event_type + u" " + message)
    except:
        raise
    else:
        signals.event_logged.send_robust(sender=entry)
        return entry


# Expose aliases for common filter functions
related_to = Entry.objects.related_to
objects_not_represented = Entry.objects.objects_not_represented
objects_represented = Entry.objects.objects_represented
filter = Entry.objects.filter
exclude = Entry.objects.exclude
all_types = Entry.objects.all_types



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-papertrail',
    version='1.1.4',
    packages=['papertrail'],
    install_requires=[
        'Django>=1.7',
        'six',
        'django-apptemplates',
    ],
    include_package_data=True,
    license='Apache Software License',
    description='An elegant solution for keeping a relational log of chronological events in a Django application.',
    long_description=README,
    url='https://www.github.com/FundersClub/django-papertrail',
    author='Eran Rundstein / FundersClub Inc.',
    author_email='eran@fundersclub.com',
    classifiers=[
        'Environment :: Web Environment',
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Database',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)
